@@ -1,4 +1,4 @@
-/*	$OpenBSD: ntpd.c,v 1.14 2004/08/12 16:33:59 henning Exp $ */
+/*	$OpenBSD: ntpd.c,v 1.25 2004/12/06 20:57:17 mickey Exp $ */
 
 /*
  * Copyright (c) 2003, 2004 Henning Brauer <henning@openbsd.org>
@@ -37,14 +37,14 @@ void	sighdlr(int);
 void	usage(void);
 int	main(int, char *[]);
 int	check_child(pid_t, const char *);
-int	dispatch_imsg(void);
+int	dispatch_imsg(struct ntpd_conf *);
 void	ntpd_adjtime(double);
+void	ntpd_settime(double);
 
-int			rfd = -1;
-volatile sig_atomic_t	quit = 0;
-volatile sig_atomic_t	reconfig = 0;
-volatile sig_atomic_t	sigchld = 0;
-struct imsgbuf		ibuf;
+volatile sig_atomic_t	 quit = 0;
+volatile sig_atomic_t	 reconfig = 0;
+volatile sig_atomic_t	 sigchld = 0;
+struct imsgbuf		*ibuf;
 
 void
 sighdlr(int sig)
@@ -68,7 +68,7 @@ usage(void)
 {
 	extern char *__progname;
 
-	fprintf(stderr, "usage: %s [-d] [-f file]\n", __progname);
+	fprintf(stderr, "usage: %s [-dSs] [-f file]\n", __progname);
 	exit(1);
 }
 
@@ -81,9 +81,8 @@ main(int argc, char *argv[])
 	struct ntpd_conf	 conf;
 	struct pollfd		 pfd[POLL_MAX];
 	pid_t			 chld_pid = 0, pid;
-	char			*conffile;
-	int			 debug = 0;
-	int			 ch, nfds;
+	const char		*conffile;
+	int			 ch, nfds, timeout = INFTIM;
 	int			 pipe_chld[2];
 
 	conffile = CONFFILE;
@@ -92,14 +91,20 @@ main(int argc, char *argv[])
 
 	log_init(1);		/* log to stderr until daemonized */
 
-	while ((ch = getopt(argc, argv, "df:")) != -1) {
+	while ((ch = getopt(argc, argv, "df:sS")) != -1) {
 		switch (ch) {
 		case 'd':
-			debug = 1;
+			conf.debug = 1;
 			break;
 		case 'f':
 			conffile = optarg;
 			break;
+		case 's':
+			conf.settime = 1;
+			break;
+		case 'S':
+			conf.settime = 0;
+			break;
 		default:
 			usage();
 			/* NOTREACHED */
@@ -120,10 +125,12 @@ main(int argc, char *argv[])
 	}
 	endpwent();
 
-	log_init(debug);
-
-	if (!debug)
-		daemon(1, 0);
+	if (!conf.settime) {
+		log_init(conf.debug);
+		if (!conf.debug)
+			daemon(1, 0);
+	} else
+		timeout = 15 * 1000;
 
 	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe_chld) == -1)
 		fatal("socketpair");
@@ -140,41 +147,55 @@ main(int argc, char *argv[])
 
 	close(pipe_chld[1]);
 
-	imsg_init(&ibuf, pipe_chld[0]);
+	if ((ibuf = malloc(sizeof(struct imsgbuf))) == NULL)
+		fatal(NULL);
+	imsg_init(ibuf, pipe_chld[0]);
 
 	while (quit == 0) {
-		pfd[PFD_PIPE].fd = ibuf.fd;
+		pfd[PFD_PIPE].fd = ibuf->fd;
 		pfd[PFD_PIPE].events = POLLIN;
-		if (ibuf.w.queued)
+		if (ibuf->w.queued)
 			pfd[PFD_PIPE].events |= POLLOUT;
 
-		if ((nfds = poll(pfd, 1, INFTIM)) == -1)
+		if ((nfds = poll(pfd, 1, timeout)) == -1)
 			if (errno != EINTR) {
 				log_warn("poll error");
 				quit = 1;
 			}
 
+		if (nfds == 0 && conf.settime) {
+			log_debug("no reply received, skipping initial time"
+			    "setting");
+			conf.settime = 0;
+			timeout = INFTIM;
+			log_init(conf.debug);
+			if (!conf.debug)
+				daemon(1, 0);
+		}
+
 		if (nfds > 0 && (pfd[PFD_PIPE].revents & POLLOUT))
-			if (msgbuf_write(&ibuf.w) < 0) {
+			if (msgbuf_write(&ibuf->w) < 0) {
 				log_warn("pipe write error (to child");
 				quit = 1;
 			}
 
 		if (nfds > 0 && pfd[PFD_PIPE].revents & POLLIN) {
 			nfds--;
-			if (dispatch_imsg() == -1)
+			if (dispatch_imsg(&conf) == -1)
 				quit = 1;
 		}
 
 		if (sigchld) {
-			if (check_child(chld_pid, "child"))
+			if (check_child(chld_pid, "child")) {
 				quit = 1;
+				chld_pid = 0;
+			}
 			sigchld = 0;
 		}
 
 	}
 
-	signal(SIGCHLD, SIG_IGN);
+	signal(SIGCHLD, SIG_DFL);
 
 	if (chld_pid)
 		kill(chld_pid, SIGTERM);
@@ -185,6 +206,8 @@ main(int argc, char *argv[])
 			fatal("wait");
 	} while (pid != -1 || (pid == -1 && errno == EINTR));
 
+	msgbuf_clear(&ibuf->w);
+	free(ibuf);
 	log_info("Terminating");
 	return (0);
 }
@@ -210,7 +233,7 @@ check_child(pid_t pid, const char *pname)
 }
 
 int
-dispatch_imsg(void)
+dispatch_imsg(struct ntpd_conf *conf)
 {
 	struct imsg		 imsg;
 	int			 n, cnt;
@@ -219,7 +242,7 @@ dispatch_imsg(void)
 	struct ntp_addr		*h, *hn;
 	struct buf		*buf;
 
-	if ((n = imsg_read(&ibuf)) == -1)
+	if ((n = imsg_read(ibuf)) == -1)
 		return (-1);
 
 	if (n == 0) {	/* connection closed */
@@ -228,7 +251,7 @@ dispatch_imsg(void)
 	}
 
 	for (;;) {
-		if ((n = imsg_get(&ibuf, &imsg)) == -1)
+		if ((n = imsg_get(ibuf, &imsg)) == -1)
 			return (-1);
 
 		if (n == 0)
@@ -241,20 +264,33 @@ dispatch_imsg(void)
 			memcpy(&d, imsg.data, sizeof(d));
 			ntpd_adjtime(d);
 			break;
+		case IMSG_SETTIME:
+			if (imsg.hdr.len != IMSG_HEADER_SIZE + sizeof(d))
+				fatal("invalid IMSG_SETTIME received");
+			if (!conf->settime)
+				break;
+			memcpy(&d, imsg.data, sizeof(d));
+			ntpd_settime(d);
+			/* daemonize now */
+			log_init(conf->debug);
+			if (!conf->debug)
+				daemon(1, 0);
+			conf->settime = 0;
+			break;
 		case IMSG_HOST_DNS:
 			name = imsg.data;
 			if (imsg.hdr.len != strlen(name) + 1 + IMSG_HEADER_SIZE)
 				fatal("invalid IMSG_HOST_DNS received");
 			if ((cnt = host_dns(name, &hn)) > 0) {
-				buf = imsg_create(&ibuf, IMSG_HOST_DNS,
-				    imsg.hdr.peerid,
+				buf = imsg_create(ibuf, IMSG_HOST_DNS,
+				    imsg.hdr.peerid, 0,
 				    cnt * sizeof(struct sockaddr_storage));
 				if (buf == NULL)
 					break;
 				for (h = hn; h != NULL; h = h->next) {
 					imsg_add(buf, &h->ss, sizeof(h->ss));
 				}
-				imsg_close(&ibuf, buf);
+				imsg_close(ibuf, buf);
 			}
 			break;
 		default:
@@ -270,8 +306,39 @@ ntpd_adjtime(double d)
 {
 	struct timeval	tv;
 
+	if (d >= (double)LOG_NEGLIGEE / 1000)
+		log_info("adjusting local clock by %fs", d);
+	else
+		log_debug("adjusting local clock by %fs", d);
 	d_to_tv(d, &tv);
-	log_info("adjusting local clock by %fs", d);
 	if (adjtime(&tv, NULL) == -1)
 		log_warn("adjtime failed");
 }
+
+void
+ntpd_settime(double d)
+{
+	struct timeval	tv, curtime;
+	char		buf[80];
+	time_t		tval;
+
+	/* if the offset is small, don't call settimeofday */
+	if (d < SETTIME_MIN_OFFSET && d > -SETTIME_MIN_OFFSET)
+		return;
+
+	d_to_tv(d, &tv);
+	if (gettimeofday(&curtime, NULL) == -1)
+		log_warn("gettimeofday");
+	curtime.tv_sec += tv.tv_sec;
+	curtime.tv_usec += tv.tv_usec;
+	if (curtime.tv_usec > 1000000) {
+		curtime.tv_sec++;
+		curtime.tv_usec -= 1000000;
+	}
+	if (settimeofday(&curtime, NULL) == -1)
+		log_warn("settimeofday");
+	tval = curtime.tv_sec;
+	strftime(buf, sizeof(buf), "%a %b %e %H:%M:%S %Z %Y",
+	    localtime(&tval));
+	log_info("set local clock to %s (offset %fs)", buf, d);
+}
